from typing import TypeVar

AgentControllerConfig = TypeVar("AgentControllerConfig")
AgentControllerData = TypeVar("AgentControllerData")
StateMetrics = TypeVar("StateMetrics")
ActionAssociatedLearningData = TypeVar("ActionAssociatedLearningData")
